from pathlib import Path
from shutil import copy2
import gzip
import oxipng
import pytest


@pytest.fixture
def infile(tmpdir):
    filename = "shipitsquirrel.png"
    orig_path = Path() / "test" / filename
    input_path = Path(tmpdir) / filename
    copy2(orig_path, input_path)
    return input_path


@pytest.fixture
def outfile(tmpdir):
    return Path(tmpdir) / "shipitsquirrel-optimized.png"


@pytest.fixture(scope="session")
def indata():
    with open(Path() / "test" / "shipitsquirrel.png", "rb") as f:
        return f.read()


@pytest.fixture(scope="session")
def rawdata():
    with gzip.GzipFile(Path() / "test" / "python-logo-only.raw.gz") as f:
        return f.read()


def test_init(outfile: Path):
    assert not outfile.exists()


def test_optimize(infile, outfile):
    oxipng.optimize(infile, outfile)
    assert outfile.exists()
    assert infile.stat().st_size > outfile.stat().st_size


def test_optimize_level(infile, outfile):
    oxipng.optimize(infile, outfile, level=6)
    assert outfile.exists()
    assert infile.stat().st_size > outfile.stat().st_size


def test_optimize_opts(infile):
    initial_size = infile.stat().st_size
    oxipng.optimize(
        infile,
        fix_errors=True,
        force=True,
        # NOTE: set args deprecated in v9.1, will change to sequence
        filter={oxipng.RowFilter.Sub, oxipng.RowFilter.Up, oxipng.RowFilter.Average},  # type: ignore
        interlace=oxipng.Interlacing.Adam7,
        optimize_alpha=True,
        bit_depth_reduction=False,
        palette_reduction=False,
        grayscale_reduction=False,
        idat_recoding=False,
        scale_16=True,
        strip=oxipng.StripChunks.strip([b"cICP", b"sRGB"]),
        deflate=oxipng.Deflaters.libdeflater(12),
        fast_evaluation=False,
        timeout=100,
    )
    assert infile.stat().st_size != initial_size


def test_optimize_inplace(infile: Path):
    orig_size = infile.stat().st_size
    oxipng.optimize(infile)
    assert infile.stat().st_size < orig_size


def test_optimize_from_memory(indata):
    assert len(oxipng.optimize_from_memory(indata)) < len(indata)


def test_raises_pngerror():
    with pytest.raises(oxipng.PngError):
        oxipng.optimize_from_memory(b"Hello World!")


def test_raises_typeerror(indata):
    with pytest.raises(TypeError):
        oxipng.optimize_from_memory(indata, filter={1: 2})  # type: ignore


def test_strip_chunks():
    assert oxipng.StripChunks.none()
    assert oxipng.StripChunks.strip([b"sRGB"])
    assert oxipng.StripChunks.safe()
    assert oxipng.StripChunks.keep([b"sRGB", b"pHYs"])

    with pytest.raises(TypeError):
        assert oxipng.StripChunks.strip(["sRGB", 42])  # type: ignore

    with pytest.raises(TypeError):
        assert oxipng.StripChunks.keep([b"RGB"])

    with pytest.raises(TypeError):
        assert oxipng.StripChunks.keep([b"RGB123"])


def test_deflate_zopfli():
    assert oxipng.Deflaters.zopfli(1)
    assert oxipng.Deflaters.zopfli(42)
    assert oxipng.Deflaters.zopfli(255)

    with pytest.raises(TypeError):
        oxipng.Deflaters.zopfli(0)

    with pytest.raises(OverflowError):
        oxipng.Deflaters.zopfli(256)


def test_raw_image(rawdata):
    raw = oxipng.RawImage(rawdata, 269, 326)
    raw.add_png_chunk(b"sRBG", b"\0")
    raw.add_icc_profile(b"Color LCD")
    optimized = raw.create_optimized_png(
        level=2,
        fix_errors=True,
        interlace=oxipng.Interlacing.Adam7,
    )
    assert len(optimized) < len(rawdata)


def test_raw_image_rgb():
    raw = oxipng.RawImage(
        b"\1\2\3\4\5\6\7\6\5\1\2\3",
        2,
        2,
        color_type=oxipng.ColorType.rgb((4, 5, 6)),
        bit_depth=8,
    )
    assert raw.create_optimized_png()


def test_color_type():
    assert oxipng.ColorType.grayscale()
    assert oxipng.ColorType.grayscale(42)

    with pytest.raises(OverflowError):
        oxipng.ColorType.grayscale(1_000_000)

    assert oxipng.ColorType.rgb()
    assert oxipng.ColorType.rgb((0, 0, 0))
    assert oxipng.ColorType.rgb([65535, 65535, 65535])

    with pytest.raises(OverflowError):
        assert oxipng.ColorType.rgb((65535, 65536, 65535))

    assert oxipng.ColorType.indexed([[1, 2, 3, 4]])
    assert oxipng.ColorType.indexed([(i, i, i, 255) for i in range(256)])

    with pytest.raises(OverflowError):
        assert oxipng.ColorType.indexed([(255, 255, 255, 256)])

    with pytest.raises(ValueError):
        assert oxipng.ColorType.indexed([(255, 255, 255)])

    assert oxipng.ColorType.grayscale_alpha()
    assert oxipng.ColorType.rgba()



"""
Python wrapper for multithreaded .png image file optimizer oxipng
(https://github.com/shssoichiro/oxipng - written in Rust). Use this module to
reduce the file size of your PNG images.
"""

from typing import Optional, Union, Sequence
from enum import Enum
from os import PathLike

StrOrBytesPath = Union[str, bytes, PathLike]

class PngError(Exception):
    """
    Raised by optimize functions when an error is encountered while optimizing PNG files
    """

    ...

class RowFilter(Enum):
    """
    enum entries for filter option
    """

    NoOp = ...
    Sub = ...
    Up = ...
    Average = ...
    Paeth = ...
    MinSum = ...
    Entropy = ...
    Bigrams = ...
    BigEnt = ...
    Brute = ...

class Interlacing(Enum):
    """
    enum entries for interlace option
    """

    Off = ...
    Adam7 = ...

class StripChunks:
    """
    Initialization class for strip option
    """

    @staticmethod
    def none() -> "StripChunks": ...
    @staticmethod
    def strip(val: Sequence[bytes]) -> "StripChunks": ...
    @staticmethod
    def safe() -> "StripChunks": ...
    @staticmethod
    def keep(val: Sequence[bytes]) -> "StripChunks": ...
    @staticmethod
    def all() -> "StripChunks": ...

class Deflaters:
    """
    Initialization class for deflate option
    """

    @staticmethod
    def libdeflater(compression: int) -> "Deflaters": ...
    @staticmethod
    def zopfli(iterations: int) -> "Deflaters": ...

class Zopfli:
    """
    Initialize a Zopfli deflate configuration option value
    """

    def __init__(self, iterations: int) -> None: ...

class Libdeflater:
    """
    Initialize a Libdeflater deflate configuration option value
    """

    def __init__(self) -> None: ...

class ColorType:
    """
    Initialization class for RawImage color_type option
    """

    @staticmethod
    def grayscale(transparent_shade: Optional[int] = None) -> "ColorType":
        """
        Grayscale, with one color channel.
        """
        ...

    @staticmethod
    def rgb(transparent_color: Optional[Sequence[int]] = None) -> "ColorType":
        """
        RGB, with three color channels. Specify optional color value that should
        be rendered as transparent.
        """
        ...

    @staticmethod
    def indexed(palette: Sequence[Sequence[int]]) -> "ColorType":
        """
        Indexed, with one byte per pixel representing a color from the palette.
        Specify palette containing the colors used, up to 256 entries
        """
        ...

    @staticmethod
    def grayscale_alpha() -> "ColorType":
        """
        Grayscale + Alpha, with two color channels.
        """
        ...

    @staticmethod
    def rgba() -> "ColorType":
        """
        RGBA, with four color channels.
        """
        ...

class RawImage:
    """
    Create an optimized PNG file from raw image data
    """

    def __init__(
        self,
        data: Union[bytes, bytearray],
        width: int,
        height: int,
        *,
        color_type: ColorType = ...,
        bit_depth: int = 8,
    ) -> None: ...

    def add_png_chunk(self, name: bytes, data: Union[bytes, bytearray]) -> None:
        """
        Add a png chunk, such as `b"iTXt"`, to be included in the output
        """
        ...

    def add_icc_profile(self, data: bytes) -> None:
        """
        Add an ICC profile for the image
        """
        ...

    def create_optimized_png(
        self,
        *,
        level: int = 2,
        fix_errors: bool = False,
        force: bool = False,
        filter: Sequence[RowFilter] = [RowFilter.NoOp],
        interlace: Optional[Interlacing] = None,
        optimize_alpha: bool = False,
        bit_depth_reduction: bool = True,
        color_type_reduction: bool = True,
        palette_reduction: bool = True,
        grayscale_reduction: bool = True,
        idat_recoding: bool = True,
        scale_16: bool = False,
        strip: StripChunks = StripChunks.none(),
        deflate: Deflaters = Deflaters.libdeflater(11),
        fast_evaluation: bool = False,
        timeout: Optional[int] = None,
    ) -> bytes:
        """
        Create an optimized png from the raw image data. Full option
        descriptions at https://github.com/nfrasser/pyoxipng#options
        """
        ...

def optimize(
    input: StrOrBytesPath,
    output: Optional[StrOrBytesPath] = ...,
    *,
    level: int = 2,
    fix_errors: bool = False,
    force: bool = False,
    filter: Sequence[RowFilter] = [RowFilter.NoOp],
    interlace: Optional[Interlacing] = None,
    optimize_alpha: bool = False,
    bit_depth_reduction: bool = True,
    color_type_reduction: bool = True,
    palette_reduction: bool = True,
    grayscale_reduction: bool = True,
    idat_recoding: bool = True,
    scale_16: bool = False,
    strip: StripChunks = StripChunks.none(),
    deflate: Deflaters = Deflaters.libdeflater(11),
    fast_evaluation: bool = False,
    timeout: Optional[int] = None,
) -> None:
    """
    Optimize a file on disk. Full option descriptions at
    https://github.com/nfrasser/pyoxipng#options
    """
    ...

def optimize_from_memory(
    data: bytes,
    *,
    level: int = 2,
    fix_errors: bool = False,
    force: bool = False,
    filter: Sequence[RowFilter] = [RowFilter.NoOp],
    interlace: Optional[Interlacing] = None,
    optimize_alpha: bool = False,
    bit_depth_reduction: bool = True,
    color_type_reduction: bool = True,
    palette_reduction: bool = True,
    grayscale_reduction: bool = True,
    idat_recoding: bool = True,
    scale_16: bool = False,
    strip: StripChunks = StripChunks.none(),
    deflate: Deflaters = Deflaters.libdeflater(11),
    fast_evaluation: bool = False,
    timeout: Optional[int] = None,
) -> bytes:
    """
    Optimize raw data from a PNG file loaded in Python as a bytes object. Full
    option descriptions at https://github.com/nfrasser/pyoxipng#options
    """
    ...
